"""MCP SSH Server implementation."""

import json
import logging
import os
import sys
from typing import Dict, List, Optional, Any, Iterator

import paramiko
from pydantic import BaseModel, Field
from mcp import types
from mcp.server import FastMCP

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class SSHSession:
    """SSH Session that connects to a remote server and supports SFTP file operations."""
    
    def __init__(self, hostname: str, port: int, username: str, key_filename: str):
        self.hostname = hostname
        self.port = port
        self.username = username
        self.key_filename = key_filename
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.connected = False
        self.sftp = None
    
    def connect(self) -> bool:
        try:
            self.client.connect(
                hostname=self.hostname,
                port=self.port,
                username=self.username,
                key_filename=self.key_filename
            )
            self.sftp = self.client.open_sftp()
            self.connected = True
            return True
        except Exception as e:
            logger.error(f"SSH connection error: {e}")
            return False
    
    def execute_command(self, command: str) -> Dict[str, Any]:
        if not self.connected:
            if not self.connect():
                return {"stdout": "", "stderr": "Not connected to SSH server", "exit_code": -1}
        try:
            stdin, stdout, stderr = self.client.exec_command(command)
            exit_code = stdout.channel.recv_exit_status()
            return {
                "stdout": stdout.read().decode("utf-8"),
                "stderr": stderr.read().decode("utf-8"),
                "exit_code": exit_code
            }
        except Exception as e:
            logger.error(f"Command execution error: {e}")
            return {"stdout": "", "stderr": str(e), "exit_code": -1}
    
    def close(self) -> None:
        if self.sftp:
            self.sftp.close()
            self.sftp = None
        if self.connected:
            self.client.close()
            self.connected = False
    
    # --- SFTP/SCP Methods ---
    def get_disk_usage(self, remote_path: str = ".") -> Dict[str, Any]:
        """Get disk usage on the remote system using 'df -k'."""
        result = self.execute_command(f"df -k {remote_path}")
        if result["exit_code"] != 0:
            return {"success": False, "error": result["stderr"]}
        try:
            lines = result["stdout"].splitlines()
            if len(lines) < 2:
                return {"success": False, "error": "Unexpected df output"}
            parts = lines[1].split()
            total_kb = int(parts[1])
            used_kb = int(parts[2])
            avail_kb = int(parts[3])
            return {"success": True, "total": total_kb * 1024, "used": used_kb * 1024, "avail": avail_kb * 1024}
        except Exception as e:
            return {"success": False, "error": str(e)}
    
    def list_directory(self, remote_path: str) -> Dict[str, Any]:
        if not self.sftp:
            return {"success": False, "error": "Not connected to SFTP"}
        try:
            files = self.sftp.listdir_attr(remote_path)
            return {"success": True, "files": [{"filename": f.filename, "size": f.st_size, "mtime": f.st_mtime, "mode": f.st_mode} for f in files]}
        except Exception as e:
            return {"success": False, "error": str(e)}
    
    def remove_file(self, remote_path: str) -> Dict[str, Any]:
        if not self.sftp:
            return {"success": False, "error": "Not connected to SFTP"}
        try:
            self.sftp.remove(remote_path)
            return {"success": True}
        except Exception as e:
            return {"success": False, "error": str(e)}
    
    def upload_file(self, local_path: str, remote_path: str, max_size: int = 1073741824, on_conflict: str = "FAIL") -> Dict[str, Any]:
        import os, time
        if not self.sftp:
            return {"success": False, "error": "Not connected to SFTP"}
        try:
            local_size = os.stat(local_path).st_size
            if local_size > max_size:
                return {"success": False, "error": f"File too large: {local_size} bytes > {max_size} bytes"}
            disk = self.get_disk_usage(os.path.dirname(remote_path) or ".")
            if not disk.get("success"):
                return {"success": False, "error": "Failed to check remote disk space: " + disk.get("error", "")}
            if disk["avail"] < local_size * 2:
                return {"success": False, "error": f"Insufficient remote disk space: need {local_size*2} bytes, have {disk['avail']} bytes"}
            # Conflict resolution
            try:
                self.sftp.stat(remote_path)
                file_exists = True
            except FileNotFoundError:
                file_exists = False
            if file_exists:
                if on_conflict == "FAIL":
                    return {"success": False, "error": "Remote file exists"}
                elif on_conflict == "OVERWRITE":
                    pass  # allow overwrite
                elif on_conflict == "RENAME":
                    base, ext = os.path.splitext(remote_path)
                    timestamp = time.strftime("(%Y%m%d%H%M%S)")
                    remote_path = f"{base}{timestamp}{ext}"
            # Upload
            try:
                self.sftp.put(local_path, remote_path)
                return {"success": True, "remote_path": remote_path, "bytes_transferred": local_size}
            except Exception as e:
                # Cleanup partial file
                try:
                    self.sftp.remove(remote_path)
                except Exception:
                    pass
                return {"success": False, "error": f"Transfer failed: {e}"}
        except Exception as e:
            return {"success": False, "error": str(e)}
    
    def download_file(self, remote_path: str, local_path: str, max_size: int = 1073741824, on_conflict: str = "FAIL") -> Dict[str, Any]:
        import os, time
        if not self.sftp:
            return {"success": False, "error": "Not connected to SFTP"}
        try:
            fileinfo = self.sftp.stat(remote_path)
            remote_size = fileinfo.st_size
            if remote_size > max_size:
                return {"success": False, "error": f"Remote file too large: {remote_size} bytes > {max_size} bytes"}
            st = os.statvfs(os.path.dirname(local_path) or ".")
            local_free = st.f_bavail * st.f_frsize
            if local_free < remote_size * 2:
                return {"success": False, "error": f"Insufficient local disk space: need {remote_size*2} bytes, have {local_free} bytes"}
            # Conflict resolution
            if os.path.exists(local_path):
                if on_conflict == "FAIL":
                    return {"success": False, "error": "Local file exists"}
                elif on_conflict == "OVERWRITE":
                    pass
                elif on_conflict == "RENAME":
                    base, ext = os.path.splitext(local_path)
                    timestamp = time.strftime("(%Y%m%d%H%M%S)")
                    local_path = f"{base}{timestamp}{ext}"
            # Download
            try:
                self.sftp.get(remote_path, local_path)
                return {"success": True, "local_path": local_path, "bytes_transferred": remote_size}
            except Exception as e:
                # Cleanup partial file
                try:
                    os.remove(local_path)
                except Exception:
                    pass
                return {"success": False, "error": f"Download failed: {e}"}
        except Exception as e:
            return {"success": False, "error": str(e)}

    """SSH Session that connects to a remote server."""
    
    def __init__(self, hostname: str, port: int, username: str, key_filename: str):
        """Initialize SSH session.
        
        Args:
            hostname: Remote server hostname
            port: SSH port
            username: SSH username
            key_filename: Path to SSH key file
        """
        self.hostname = hostname
        self.port = port
        self.username = username
        self.key_filename = key_filename
        self.client = paramiko.SSHClient()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.connected = False
        
    def connect(self) -> bool:
        """Connect to SSH server.
        
        Returns:
            True if connection successful, False otherwise
        """
        try:
            self.client.connect(
                hostname=self.hostname,
                port=self.port,
                username=self.username,
                key_filename=self.key_filename
            )
            self.connected = True
            return True
        except Exception as e:
            logger.error(f"SSH connection error: {e}")
            return False
            
    def execute_command(self, command: str) -> Dict[str, Any]:
        """Execute a command on the remote server.
        
        Args:
            command: Command to execute
            
        Returns:
            Dictionary with stdout, stderr, and exit_code
        """
        if not self.connected:
            if not self.connect():
                return {"stdout": "", "stderr": "Not connected to SSH server", "exit_code": -1}
        
        try:
            stdin, stdout, stderr = self.client.exec_command(command)
            exit_code = stdout.channel.recv_exit_status()
            
            return {
                "stdout": stdout.read().decode("utf-8"),
                "stderr": stderr.read().decode("utf-8"),
                "exit_code": exit_code
            }
        except Exception as e:
            logger.error(f"Command execution error: {e}")
            return {"stdout": "", "stderr": str(e), "exit_code": -1}
    
    def close(self) -> None:
        """Close SSH connection."""
        if self.connected:
            self.client.close()
            self.connected = False


class SSHConnectionParams(BaseModel):
    """Parameters for SSH connection."""
    
    hostname: Optional[str] = Field(None, description="SSH server hostname or IP address")
    port: Optional[int] = Field(None, description="SSH server port")
    username: Optional[str] = Field(None, description="SSH username")
    key_filename: Optional[str] = Field(None, description="Path to SSH private key file")


class FileTransferParams(BaseModel):
    local_path: str
    remote_path: str
    max_size: Optional[int] = 1073741824  # 1GB default
    on_conflict: str = "FAIL"  # FAIL, OVERWRITE, RENAME

class ListDirParams(BaseModel):
    remote_path: str

class RemoveFileParams(BaseModel):
    remote_path: str

class DiskUsageParams(BaseModel):
    remote_path: Optional[str] = "."

class CommandParams(BaseModel):
    """Parameters for executing a command."""
    
    command: str = Field(..., description="Command to execute on remote server")


class SSHServerMCP(FastMCP):
    """MCP server that provides SSH and SFTP access to remote systems."""
    
    def __init__(self, hostname=None, port=None, username=None, key_filename=None, server_name=None, tool_prefix=None):
        # Get server name from environment variable or parameter
        server_name = server_name or os.environ.get("MCP_SSH_SERVER_NAME")
        # Get configuration from environment variables or passed parameters
        self.default_hostname = hostname or os.environ.get("MCP_SSH_HOSTNAME")
        self.default_port = port or int(os.environ.get("MCP_SSH_PORT", 22))
        self.default_username = username or os.environ.get("MCP_SSH_USERNAME")
        self.default_key_filename = key_filename or os.environ.get("MCP_SSH_KEY_FILENAME")
        self.tool_prefix = tool_prefix or os.environ.get("MCP_SSH_TOOL_PREFIX", "")
        connect_name = f"{self.tool_prefix}ssh_connect"
        execute_name = f"{self.tool_prefix}ssh_execute"
        disconnect_name = f"{self.tool_prefix}ssh_disconnect"
        upload_name = f"{self.tool_prefix}scp_upload"
        download_name = f"{self.tool_prefix}scp_download"
        listdir_name = f"{self.tool_prefix}scp_listdir"
        remove_name = f"{self.tool_prefix}scp_remove"
        diskusage_name = f"{self.tool_prefix}scp_disk_usage"
        self.add_tool(self.ssh_connect, name=connect_name, description=f"Connect to SSH server: {server_name}")
        self.add_tool(self.ssh_execute, name=execute_name, description=f"Execute a command on SSH server: {server_name}")
        self.add_tool(self.ssh_disconnect, name=disconnect_name, description=f"Disconnect from SSH server: {server_name}")
        self.add_tool(self.scp_upload, name=upload_name, description=f"Upload a file via SCP/SFTP: {server_name}")
        self.add_tool(self.scp_download, name=download_name, description=f"Download a file via SCP/SFTP: {server_name}")
        self.add_tool(self.scp_listdir, name=listdir_name, description=f"List remote directory via SFTP: {server_name}")
        self.add_tool(self.scp_remove, name=remove_name, description=f"Remove remote file via SFTP: {server_name}")
        self.add_tool(self.scp_disk_usage, name=diskusage_name, description=f"Get remote disk usage via SFTP: {server_name}")
        self.ssh_session = None
    
    def ssh_connect(self, params: SSHConnectionParams) -> Dict[str, Any]:
        hostname = self.default_hostname
        username = self.default_username
        key_filename = self.default_key_filename
        port = self.default_port if self.default_port else params.port
        if not all([hostname, username, key_filename]):
            missing = []
            if not hostname: missing.append("hostname")
            if not username: missing.append("username")
            if not key_filename: missing.append("key_filename")
            return {"success": False, "message": f"Missing required parameters: {', '.join(missing)}"}
        self.ssh_session = SSHSession(
            hostname=hostname,
            port=port,
            username=username,
            key_filename=key_filename
        )
        if self.ssh_session.connect():
            return {"success": True, "message": f"Connected to {hostname}"}
        else:
            return {"success": False, "message": "Failed to connect to SSH server"}
    
    def ssh_execute(self, params: CommandParams) -> Dict[str, Any]:
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        return self.ssh_session.execute_command(params.command)
    
    def ssh_disconnect(self) -> Dict[str, Any]:
        if not self.ssh_session:
            return {"success": True, "message": "Not connected to SSH server"}
        self.ssh_session.close()
        return {"success": True, "message": "Disconnected from SSH server"}
    
    # --- SFTP/SCP Tool Methods ---
    def scp_upload(self, params: FileTransferParams) -> Dict[str, Any]:
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        return self.ssh_session.upload_file(
            local_path=params.local_path,
            remote_path=params.remote_path,
            max_size=params.max_size or 1073741824,
            on_conflict=params.on_conflict
        )

    def scp_download(self, params: FileTransferParams) -> Dict[str, Any]:
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        return self.ssh_session.download_file(
            remote_path=params.remote_path,
            local_path=params.local_path,
            max_size=params.max_size or 1073741824,
            on_conflict=params.on_conflict
        )

    def scp_listdir(self, params: ListDirParams) -> Dict[str, Any]:
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        return self.ssh_session.list_directory(params.remote_path)

    def scp_remove(self, params: RemoveFileParams) -> Dict[str, Any]:
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        return self.ssh_session.remove_file(params.remote_path)

    def scp_disk_usage(self, params: DiskUsageParams) -> Dict[str, Any]:
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        return self.ssh_session.get_disk_usage(params.remote_path)

    """MCP server that provides SSH access to remote systems."""
    
    def __init__(self, hostname=None, port=None, username=None, key_filename=None, server_name=None, tool_prefix=None):
        """Initialize SSH server.
        
        Args:
            hostname: SSH server hostname from environment or config
            port: SSH server port from environment or config
            username: SSH username from environment or config
            key_filename: Path to SSH key file from environment or config
            server_name: Custom name for the server instance (for multiple servers)
            tool_prefix: Prefix for tool names (e.g., 'server1_' for 'server1_ssh_connect')
        """
        # Get server name from environment variable or parameter
        server_name = server_name or os.environ.get("MCP_SSH_SERVER_NAME", "SSH Server")
        super().__init__(name=server_name)
        self.ssh_session: Optional[SSHSession] = None
        
        # Get configuration from environment variables or passed parameters
        self.default_hostname = hostname or os.environ.get("MCP_SSH_HOSTNAME")
        self.default_port = port or int(os.environ.get("MCP_SSH_PORT", 22))
        self.default_username = username or os.environ.get("MCP_SSH_USERNAME")
        self.default_key_filename = key_filename or os.environ.get("MCP_SSH_KEY_FILENAME")
        
        # Get tool prefix from parameter or environment variable
        self.tool_prefix = tool_prefix or os.environ.get("MCP_SSH_TOOL_PREFIX", "")
        
        # Register tools with optional prefix
        connect_name = f"{self.tool_prefix}ssh_connect"
        execute_name = f"{self.tool_prefix}ssh_execute"
        disconnect_name = f"{self.tool_prefix}ssh_disconnect"
        
        self.add_tool(self.ssh_connect, name=connect_name, description=f"Connect to SSH server: {server_name}")
        self.add_tool(self.ssh_execute, name=execute_name, description=f"Execute a command on SSH server: {server_name}")
        self.add_tool(self.ssh_disconnect, name=disconnect_name, description=f"Disconnect from SSH server: {server_name}")
    
    def ssh_connect(self, params: SSHConnectionParams) -> Dict[str, Any]:
        """Connect to an SSH server.
        
        Args:
            params: SSH connection parameters (ignored for security-critical fields)
            
        Returns:
            Result of connection attempt
        """
        # Always use config/environment variables for security-critical fields
        hostname = self.default_hostname
        username = self.default_username
        key_filename = self.default_key_filename
        
        # Only allow port to be optionally specified in the request
        port = self.default_port if self.default_port else params.port
        
        # Validate that we have all required parameters
        if not all([hostname, username, key_filename]):
            missing = []
            if not hostname: missing.append("hostname")
            if not username: missing.append("username")
            if not key_filename: missing.append("key_filename")
            return {"success": False, "message": f"Missing required parameters: {', '.join(missing)}"}
        
        self.ssh_session = SSHSession(
            hostname=hostname,
            port=port,
            username=username,
            key_filename=key_filename
        )
        
        if self.ssh_session.connect():
            return {"success": True, "message": f"Connected to {hostname}"}
        else:
            return {"success": False, "message": "Failed to connect to SSH server"}
    
    def ssh_execute(self, params: CommandParams) -> Dict[str, Any]:
        """Execute a command on the SSH server.
        
        Args:
            params: Command parameters
            
        Returns:
            Command execution result
        """
        if not self.ssh_session or not self.ssh_session.connected:
            return {"success": False, "message": "Not connected to SSH server"}
        
        return self.ssh_session.execute_command(params.command)
    
    def ssh_disconnect(self) -> Dict[str, Any]:
        """Disconnect from the SSH server.
        
        Returns:
            Disconnect result
        """
        if not self.ssh_session:
            return {"success": True, "message": "Not connected to SSH server"}
        
        self.ssh_session.close()
        return {"success": True, "message": "Disconnected from SSH server"}


def main():
    """Run the MCP SSH server."""
    # Get custom server name and tool prefix from environment if specified
    server_name = os.environ.get("MCP_SSH_SERVER_NAME")
    tool_prefix = os.environ.get("MCP_SSH_TOOL_PREFIX")
    
    # Create server with config from environment variables
    server = SSHServerMCP(
        server_name=server_name,
        tool_prefix=tool_prefix
    )
    
    # Run the server with stdio transport
    server.run(transport="stdio")


if __name__ == "__main__":
    main()